% electrodeMasker(imageMatrix,params)
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Notes ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% -default values are based on the asumptions that intensity values range
% between 0-255 with the electrode as the 'darkest' region.
%
% -This is a wraper function consiting of 'masker', and 'imgFilterF'.
%
% -This function filters and "masks" the image twice, consecutively in
% order to achiev better results then a single filter and "masking" the 
% image. As such there is options in 'params' to set both blurFactor/stD 
% and threshold constants, for each "filter & mask run". using 
% 'params.run1.blurFactor' & 'params.run1.threshold' for run1, and 
% 'params.run2.blurFactor' & 'params.run2.threshold' for run2.
%
% -The input of 'params.run2.threshold' is divided by a constant 'DivConst'  
% (set at 255). This is done purely so the numbers input for run1.threshold
% and run2.threshold use "comparable" numbers. Without modification,
% 'params.run2.threshold' would need to be between 0.00 and ~0.99.
% This input modifier can be switched off by setting 'params.inputMod' to
% false.
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% =============================== Input ==================================
% _____Required_____
%
% imageMatrix:
%   -2D matrix of image containing an electrode. intensity values should be
%   between 0-255 for the function to work properly.
% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% _____Optional(params)_____
%
% params.run1.blurFactor: number ~ default = 2;
% params.run2.blurFactor: number ~ default = 10;
%   -Used as the params.deviation argument for imgFilterF. 
% params.run1.threshold: number ~ default = 150;
% params.run2.threshold: number ~ default = 40;
%   -Used as the cuttOff argument for masker.
% params.inverse: bool ~ default = true;
%   -For a grescale image with intensity values between 0-255; true will
%   result in an inverse in intensity values. low-->high/high-->low.
% params.logical: bool ~ default = true;
%   -If true, the output is a logic/bool 01 matrix, if false then the output
%   output is a numerical 01 matrix.
% params.inputMod: bool ~ default = true;
%   -If true, 'params.run2.threshold' input is internaly divided by a 
%   constnt (255). If false no modification is carried out. 
% =============================== Output =================================
% electrodeMask:
%   -logical (default) or numerical 01 matrix. Elements covering the area 
%   corrospondting to the location of the electrode in the input image 
%   are set to 1. All other ellements are set to 0. 
% ========================================================================

function electrodeMask = electrodeMasker(imageMatrix,params)

DivConst = 255;
% DivConst: -the input of run2.threshold is divided by this number.  
%           This is done purely so the numbers input for run1.threshold
%           and run2.threshold use "comparable" numbers (if not run2.
%           threshold would need to be between 0.00 and ~0.99).


% -------- Defaults --------
divConstMod = true;
invert = true;
logic = true;
run1 = struct;
run1.stD = 2;
run1.threshold = 150;
run2 = struct;
run2.stD = 10;
run2.threshold = 40/DivConst;
% --------------------------

if nargin > 1
    if isfield(params,'inverse')
        invert = params.inverse;
    end
    if isfield(params,'logical')
        logic = params.logical;
    end
    if isfield(params,'inputMod')
        divConstMod = params.inputMod;
        if ~divConstMod
            DivConst = 1;
        end
    end
    if isfield(params,'run1')
        if isfield(params.run1,'threshold')
            run1.threshold = params.run1.threshold;
        end
        if isfield(params.run1,'blurFactor')
            run1.stD = params.run1.blurFactor;
        end
    end
    if isfield(params,'run2')
        if isfield(params.run2,'threshold')
            run2.threshold = params.run2.threshold/DivConst;
        end
        if isfield(params.run2,'blurFactor')
            run2.stD = params.run2.blurFactor;
        end
    end
end



imgA = imageMatrix;
if size(imgA,3)==3
    imgA = rgb2gray(imgA);
end
if invert
    imgA = 255 - imgA;
end

paramy1.pad = 'reflection';
paramy1.quiet = true;
paramy1.deviation = run1.stD;
%-------------------------- imgFilterF()
imgB = imgFilterF(imgA,'gaussian','fourier',paramy1);
%--------------------------
%-------------------------- masker()
mask1 = masker(imgB,run1.threshold);
%--------------------------
paramy2.pad = 'reflection';
paramy2.quiet = true;
paramy2.deviation = run2.stD;
%-------------------------- imgFilterF()
imgC = imgFilterF(mask1,'gaussian','fourier',paramy2);
%--------------------------
%-------------------------- masker()
mask2 = masker(imgC,run2.threshold);
%--------------------------

if logic
    electrodeMask = logical(mask2);
else
    electrodeMask = mask2;
end

