% getLocations(targetParameters, imageFolder, maskX, params)
%
%
% <><><><><><> Required Functions <><><><><> %
% > getGrayscale
% > interExcludeCombiner
% > excludeMasked
% > findObjectsX
% <><><><><><><><><><><><><><><><><><><><><> %
% ==== Input ==== % 
% --- Required --- %
%
% targetParameters: - cell array with 5-7 columns and variable rows.
%                   - each row represents a seperate parameter set used in the
%                     findObjectX function.
%        [required] - Column 1 = filterA deviation (lowerBound length) <2>.
%        [required] - Column 2 = filterB deviation (upperBound length) <3>.
%        [required] - Column 3 = intensity threshold <4.5>.
%        [required] - Column 4 = exclusion length (clossest distance in pixels 
%                     2 points found using the same parameters can be) <20>.
%        [required] - Column 5 = invert (i.e. 255 - image) <false>.
%
% imageFolder: - address of folder containing image files.
%
% --- optional --- %
%
% NOTE: important to set "params.interExclude.exclusionLength"
%
% maskX: -bool/01 matrix with 1 representing blocked pixels and 0 for
%        accesible pixels
%
% params.interExclude.exclusionLength: minimum number of pixels between points from different
%                    sets allowed. if 2 points close to each other, the 
%                    brightest will be selected.
%                    > Default = roughly* 20 pixels.
%
% params.interExclude.useIntensity: default variies depending on if the are > 2 columns
%                      > If set to false, will not use intensity even if
%                        available.
%
% params.interExclude.prioritiseA: default is true (i.e. prioritise A over B). if false 
%                     setB is prioritised over set A.
%
% params.interExclude.accurate: default is false.
%                  > false --> use faster, less acurate measuring method. 
%                  > true --> use slower, more acurate measuring method.
%
% params.interExclude.giveBackNormalisedIntensity: 
%                                 > default is true --> Intensities are 
%                                   normalised and the output combined list
%                                   has normalised intensities.
%                                 > If false --> Intensities pass through
%                                   unchanged, but normalisation is 
%                                   used during comparing intensities.
%
% params.quiet: default is true --> supress diplaying of processing time. 
%               > If false --> diplays processing time.
%
% params.startIndex: > instead of processing from the start, will process
%                      from the given index untill reaching end or input endIndex.
%
% params.endIndex: > instead of processing to the last file, will process
%                    up untill the given end index.
%
% ==== Output ==== % NOTE: will change, i.e. get more columns, later.
%
% positionList: 3 column cell array. - col 1 = cells each containing a frames
%                                      [X,Y,Intensity] data for all located 
%                                      positions.
%                                    - col 2 = the filr/frame number.
%                                    - col 3 = the number of possitions
%                                      located in each frame.
%
%
function [positionList] = getLocations(targetParameters, imageFolder, maskX, params)



% <><><> Defaults <><><> %

endIndex = 0;
useLastFileAsEnd = true;
startIndex = 1; % nw
useMask = false;
maskExist = false;
targetN = 1;
invRequired = false;
% --- for interExclude parameters/arguments --- %
interExUseIntensity = true;
interExPrioritiseA = true;
interExAccurate = true;
interExLength = 20;
interExGiveBackNormalised = true;
% --- for interExclude parameters/arguments END --- %
% <><><> Defaults END <><><> %

if nargin < 2
   disp('<getLocations> Error: not enough input arguments!');
   return;
end

[targRow,targCol] = size(targetParameters);

if ~iscell(targetParameters) || targCol < 5
    disp('<parameterChecker> Error: "targetParameters" must be a cell array with a minimum of 5 columns!');
   return;
end

targetN = targRow;

if nargin > 2
    if length(maskX(1,:)) > 1
        maskExist = true;
        useMask = maskX;
    end
end

if nargin > 3 % params sorting
    % --- for interExclude parameters/arguments --- %
    if isfield(params,'interExclude')
        if isstruct(params.interExclude)
        	if isfield(params.interExclude,'exclusionLength')
                interExLength = params.interExclude.exclusionLength;
            end
            if isfield(params.interExclude,'useIntensity')
                interExUseIntensity = params.interExclude.useIntensity;
            end
            if isfield(params.interExclude,'prioritiseA')
                interExPrioritiseA = params.interExclude.prioritiseA;
            end
            if isfield(params.interExclude,'accurate')
                interExAccurate = params.interExclude.accurate;
            end
            if isfield(params.interExclude,'giveBackNormalisedIntensity')
                interExGiveBackNormalised = params.interExclude.giveBackNormalisedIntensity;
            end
        end
    end
    if isfield(params,'startIndex')
        startIndex = params.startIndex;
    end
    if isfield(params,'endIndex')
        if isnumeric(params.endIndex)
            if params.endIndex > 0 && params.endIndex >= startIndex
                endIndex = params.endIndex;
                useLastFileAsEnd = false;
            end
        end
    end
    % --- for interExclude parameters/arguments END --- %
end
% --- for interExclude parameters/arguments --- %
paramsEx.useIntensity = interExUseIntensity;
paramsEx.prioritiseA = interExPrioritiseA;
paramsEx.accurate = interExAccurate;
paramsEx.giveBackNormalisedIntensity = interExGiveBackNormalised;
% --- for interExclude parameters/arguments END --- %

% ========= PEPARATION
invList = cell2mat(targetParameters(:,5));
if max(invList(:))>0 % check if any in inv (i.e. col 5) col is true
    invRequired = true;
end

[imgOne,fileExtension,numberN,fileList] = getTopImgX(imageFolder,1);


if useLastFileAsEnd
    endIndex = numberN;
end

filesUsedN = endIndex - startIndex + 1;

positionList = cell(filesUsedN,3);

% ========= PREPARATION END
for j = startIndex:endIndex
    
    imgX = getImgFast(imageFolder, fileList, j);

    % --- grey scale conversion --- %
    if size(imgX,3)==3
        imgX = rgb2gray(imgX);
    end
    % --- grey scale conversion End --- %
    
    if invRequired % only generate inverse image if know will be required!
        imgXi = 255 - imgX;
    end
    
    rawPositionTables = cell(targetN,1);
    optionsX.mask = false; % for "findObjectsX" function.
    paramListX = struct;

    for i = 1:targetN

        if targetParameters{i,5}
            searchImg = imgXi;
        else
            searchImg = imgX;
        end

        paramListX.A.deviation = targetParameters{i,1};
        paramListX.B.deviation = targetParameters{i,2};
        paramListX.threshold = targetParameters{i,3};
        paramListX.exclusionLength = targetParameters{i,4};

        rawPositionTables{i} = findObjectsX(searchImg, paramListX, optionsX);
        if maskExist
            rawPositionTables{i} = excludeMasked(rawPositionTables{i},useMask);
        end

    end

    posSets = cell(targetN,1); % unnecessary transfer of variable???
    posSets(:,1) = rawPositionTables(:,1); % unnecessary transfer of variable???
    
    posList = posSets{1,1};
    if targetN > 1 
        for i = 1:(targetN - 1)
            % interExcludeCombiner combines 2 sets into one and removes the
            % positions too close to other positions
            posList = interExcludeCombiner(posList, posSets{i+1,1}, interExLength, paramsEx);
        end
    end
    framePosN = length(posList(:,1));
    positionList{j,1} = posList;
    positionList{j,2} = j; % frameIdx 
    positionList{j,3} = framePosN; % number of positions found in that frame
end
     
% showLocationsZ(imgX, posSets);
        
end


