% getPos(targetParameters, imageFolder, maskX, params)
%
% getPos is mainly just a wraper for get locations
% getPos returns a compatible output for putting into the tracker.
%
% <><><><><><> Required Functions <><><><><> %
% > getGrayscale
% > interExcludeCombiner
% > excludeMasked
% > findObjectsX
% <><><><><><><><><><><><><><><><><><><><><> %
% ==== Input ==== % 
% --- Required --- %
%
% targetParameters: - cell array with 5-7 columns and variable rows.
%                   - each row represents a seperate parameter set used in the
%                     findObjectX function.
%        [required] - Column 1 = filterA deviation (lowerBound length) <2>.
%        [required] - Column 2 = filterB deviation (upperBound length) <3>.
%        [required] - Column 3 = intensity threshold <4.5>.
%        [required] - Column 4 = exclusion length (clossest distance in pixels 
%                     2 points found using the same parameters can be) <20>.
%        [required] - Column 5 = invert (i.e. 255 - image) <false>.
%
% imageFolder: - address of folder containing image files.
%
% --- optional --- %
%
% NOTE: > important to set "params.interExclude.exclusionLength".
% NOTE: > Under most circumstances a mask shold be provided for "maskX".
% NOTE: > Under most circumstances an appropriate excell file containing 
%         frame time information shold be provided for "frameTimeFile".
%
% frameTimeFile: excel file with > col 1 = name of image file. 
%                                > col 2 = time signiture from start in ms.
%                                > col 3 = frame number/idx starting at 1.
%                                > 1st row contains column names/titles.
%
% maskX: -bool/01 matrix with 1 representing blocked pixels and 0 for
%        accesible pixels
%
% params.interExclude.exclusionLength: minimum number of pixels between points from different
%                    sets allowed. if 2 points close to each other, the 
%                    brightest will be selected.
%                    > Default = roughly* 20 pixels.
%
% params.interExclude.useIntensity: default variies depending on if the are > 2 columns
%                      > If set to false, will not use intensity even if
%                        available.
%
% params.interExclude.prioritiseA: default is true (i.e. prioritise A over B). if false 
%                     setB is prioritised over set A.
%
% params.interExclude.accurate: default is false.
%                  > false --> use faster, less acurate measuring method. 
%                  > true --> use slower, more acurate measuring method.
%
% params.startIndex: > instead of processing from the start, will process
%                      from the given index untill reaching end or input endIndex.
%
% params.endIndex: > instead of processing to the last file, will process
%                    up untill the given end index.
%
% params.realFrameStart: > default is either 1 or "params.startIndex".
%                        > The defaults refer to the file index which is not 
%                          necessarily the same as the frame index.
%                        > set this if the file index and the frame index 
%                          start differ. 
%
% params.quiet: default is true --> supress diplaying of processing time. 
%               > If false --> diplays processing time.
%
% params.output: string to indicate the output type
%               > 'raw' > output is the output of getLocations sub function.
%               > 'standard' > output is similar to posFinder.
%                            > Default output.
%               > 'new' > same as standard but with an extra 6th column
%                         containing normalised intensities.
%
% ==== Output ==== % NOTE: will change, i.e. get more columns, later.
%
% positionList: 2 column cell array. - col 1 = cells each containing a frames
%                                      [X,Y,Intensity] data for all located 
%                                      positions.
%                                    - col 2 = the number of possitions
%                                      located in each frame.
%
%
function [positionList] = getPos(targetParameters, imageFolder, maskX, frameTimeFile, params)


if nargin < 2
    disp('<getPos> ERROR: Not enough arguments!');
    return;
end

% ********************* getLocations SECTION start ********************** %
% ======================================================================= %

% <><><> Defaults <><><> %
gLmaskX = false;
% ~
gLparams = struct;
gLparams.startIndex = 1;
gLparams.endIndex = 0;
gLparams.quiet = false;
% ~
gLparams.interExclude = struct;
gLparams.interExclude.exclusionLength = 20;
gLparams.interExclude.useIntensity = true;
gLparams.interExclude.prioritiseA = true;
gLparams.interExclude.accurate = true;
gLparams.interExclude.giveBackNormalisedIntensity = true;
% ~
frameTimesAvailable = false;


% --- passing over inputs --- %
if nargin > 2
    gLmaskX = maskX;
end

if nargin > 4 % params sorting
    if isfield(params,'quiet')
        gLparams.quiet = params.quiet;
    end
    % --- for interExclude parameters/arguments --- %
    if isfield(params,'interExclude')
        if isstruct(params.interExclude)
        	if isfield(params.interExclude,'exclusionLength')
                gLparams.interExclude.exclusionLength = params.interExclude.exclusionLength;
            end
            if isfield(params.interExclude,'useIntensity')
                gLparams.interExclude.useIntensity = params.interExclude.useIntensity;
            end
            if isfield(params.interExclude,'prioritiseA')
                gLparams.interExclude.prioritiseA = params.interExclude.prioritiseA;
            end
            if isfield(params.interExclude,'accurate')
                gLparams.interExclude.accurate = params.interExclude.accurate;
            end
        end
    end
    if isfield(params,'startIndex')
        gLparams.startIndex = params.startIndex;
    end
    if isfield(params,'endIndex')
        gLparams.endIndex = params.endIndex;
    end
    % --- for interExclude parameters/arguments END --- %
end

rawOutput = getLocations(targetParameters, imageFolder, gLmaskX, gLparams);

% rawOutput (cell array) colums: 1 = [X,Y,(intensity)*], 2 = fileIdx/frameIdx, 3 = number of positions found
%

% * if intensity was input originaly.

% *********************** getLocations SECTION end ********************** %
% ======================================================================= %

% <><><> Defaults <><><> %

rawOut = false; % set to true later if params.output == 'raw'.
newSixthOut = false; % set to true later if params.output == 'new'.
frameStart = rawOutput{1,2}; % > this gives the file index start which generaly 
%                              is the same as the frame index start but not always.


% <><><> Defaults END <><><> %

if nargin > 4 % params sorting
    if isfield(params,'output')
        if strcmp(params.output,'raw')
            rawOut = true;
        end
        if strcmp(params.output,'new')
            newSixthOut = true;
        end
    end
    if isfield(params,'realFrameStart')
        if isnumeric(params.realFrameStart)
            if params.realFrameStart > 0
                frameStart = params.realFrameStart;
            end
        end
    end
    
end

if rawOut
    positionList = rawOutput;
    return; 
end

[cellRows,cellCols] = size(rawOutput);
framesN = cellRows; 
frameEndIdxPlus1 = frameStart + framesN;
frameEndIdx = frameStart + framesN - 1;

if nargin > 3
    frameTimeTable = readtable(frameTimeFile);
    frameTimesAvailable = true;
else
    frameTimeTable = false;
end

if frameTimesAvailable
    frameTimes(:,:) = table2array(frameTimeTable(:,[3,2])); 
    % "frameTimes" col 1 = frame, ciol 2 = time(ms)
else % make a psuedo frameTimes table
    frameTimes = zeros(frameEndIdx,2); % 
    frameTimes(:,1) = 1:frameEndIdx; % 'frameIdx'
    frameTimes(:,2) = 1:frameEndIdx; % 'times' set times the same as 'frameIdx'
end

% positionCellList > colums: 1 = [X,Y,(intensity)*], 2 = fileIdx, 
%                    3 = number of positions found, 4 = frameIdx,
%                    5 = time(ms).
positionCellList = cell(framesN,5); 
positionCellList(:,1:3) = rawOutput;
positionCellList(:,4:5) = num2cell(frameTimes(frameStart:frameEndIdx,1:2)); % #1

totalPosN = sum(cell2mat(positionCellList(:,3)));

% >>> NEED TO MAKE A 6 COLUMN VERSION INCLUDING NORMALISED INTENSITIES!!! <<< %

if newSixthOut
    positionList = zeros(totalPosN,6);
else
    positionList = zeros(totalPosN,5);
end

positionList(:,4) = 1:totalPosN;
tempIdxS = 1;
tempIdxE = 1;
for i = 1:framesN
    tempIdxE = tempIdxS + positionCellList{i,3} - 1;
    positionList(tempIdxS:tempIdxE,1:2) = positionCellList{i,1}(:,1:2);
    if newSixthOut
        positionList(tempIdxS:tempIdxE,6) = positionCellList{i,1}(:,3);
    end
    positionList(tempIdxS:tempIdxE,3) = positionCellList{i,4};
    positionList(tempIdxS:tempIdxE,5) = positionCellList{i,5};
    tempIdxS = tempIdxE + 1;
end
% ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ %
% % % FrameRateTable
% % % 
% % % if FrameRateTable ~= false
% % %     frameTimeTable = readtable(FrameRateTable);
% % % end
% % % 
% % % 
% % % 
% % % for ie = startFrame:(startFrame - 1 + length(theFiles))
% % %     pos(pos(:,3) == frameTimes(ie,1),5) = frameTimes(ie,2); %%%% c{k}=Res2(Res2(:,5) == idx(k),1:4);
% % % end
% % % 
% % % end

% ###################### FUNCTION END #################### %

