% interExcludeCombiner
%
% [posList] = interExcludeCombiner(posSetA, posSetB, exclusionLength, params)
%
% EXPLANATION:
%
% <><><><><><> Required Functions <><><><><> %
% > N/A
% > N/A
% <><><><><><><><><><><><><><><><><><><><><> %
% ================ INPUTS ================ %
%
% ---- Required ---- %
% > posSetA/posSetB: numerical X,Y,intensity matrix/arrays
%                    representing XY matrix/array.
%              > numerical array may have more than 3 columns, data within
%                extra columns will passthrough into the returned matrix/array.
%              > both posSetA and posSetB should have same number and
%                type of columns.
%
% > exclusionLength: minimum number of pixels between points from different
%                    sets allowed. if 2 points close to each other, the 
%                    brightest will be selected.
%
% --- optional 1 --- %
%
% params.useIntensity: default variies depending on if the are > 2 columns
%                      > If set to false, will not use intensity even if
%                        available.
%
% Note: If using intensity to prioritise, and A and B intensity is equal,
%       the prioritised set is then prioritised.
%
% params.prioritiseA: default is true (i.e. prioritise A over B). if false 
%                     setB is prioritised over set A.
%
% params.accurate: default is false.
%                  > false --> use faster, less acurate measuring method. 
%                  > true --> use slower, more acurate measuring method.
%
% params.quiet: default is true --> supress diplaying of processing time. 
%               > If false --> diplays processing time.
%
% params.giveBackNormalisedIntensity: IMPLEMENTATION IS MESSY ! (requires substantial restructuring to implement in a clean way)
%                                 > default is false --> Intensities pass 
%                                   through unchanged, but normalisation is 
%                                   used during comparing intensities.  
%                                 > If true --> Intensities are normalised
%                                   and the output combined list has
%                                   normalised intensities.
%
% ================ OUTPUT ================ %
% posList: resulting combined list, col 1 = X, col 2 = Y, col 3:n == input col 3:n
%
%
function [posList] = interExcludeCombiner(posSetA, posSetB, exclusionLength, params)

% === Defaults === %

excludeLen = 20;
colN = 2;
intensityWanted = true;
intensityAvailable = false;
intensityUsing = false;
prioritiseA = true;
acurateSlowMethod = false;
supresDisp = true;
giveNormal = false; % not imlemented yet
% === Defaults END === %

if nargin < 2
   disp('Error: Require 2 sets to compare and combine, If trying to combare within a set, use "intraExclude"!');
   return;
end



if nargin > 2
    excludeLen = exclusionLength;
end

% --- params sorter --- %
if nargin > 3
    if isfield(params,'useIntensity')
        intensityWanted = params.useIntensity;
    end
    if isfield(params,'prioritiseA')
        prioritiseA = params.prioritiseA;
    end
    if isfield(params,'accurate')
        acurateSlowMethod = params.accurate;
    end
    if isfield(params,'quiet')
        supresDisp = params.quiet;
    end
    if isfield(params,'giveBackNormalisedIntensity')
        giveNormal = params.giveBackNormalisedIntensity;
    end
end
% --- params sorter END --- %

[Arows,Acols] = size(posSetA);
[Brows,Bcols] = size(posSetB);
Abools = true(Arows,1);
Bbools = true(Brows,1);
% check both sets have a 3rd column, thus they might contain intensities.
if Acols > 2 && Bcols > 2 
    intensityAvailable = true;
end
if intensityAvailable && intensityWanted
    intensityUsing = true;
end

% <><><><> Anonymous functions <><><><> %
% > fast refers to the aproximation of distance method which does not require
%   squaring.
% > precise refers to the method which gives the squared distance. thus the 
%   results are in accordance to comparing actual distances but without 
%   using the square root function but does obviously use the squaring function.

preciseDist = @(set1,set2,idx) ((set2(:,1) - set1(idx,1)).^2) + ((set2(:,2) - set1(idx,2)).^2);
fastDistSub = @(set1,set2,idxR,idxC) abs(set2(:,idxC) - set1(idxR,idxC));
fastDist = @(set1,set2,idx) [fastDistSub(set1,set2,idx,1),fastDistSub(set1,set2,idx,2)];

preciseInd = @(temps,excludeDist) (temps(:,1) < excludeDist); 
fastInd = @(temps,excludeDist) (temps(:,1)<excludeDist & temps(:,2)<excludeDist);

% ~~~ intensity ~~~ %

% maxMinRange = @(X) [max(X),min(X),(max(X)-min(X))]; % <-- just pointless
normFunReal = @(x,min,range) (x-min)/range; % *
normFunI = @(x,const1,const2) (x*const1)-const2; % **
normFunDud = @(x,min,range) x; % *
% <><><><> Anonymous functions END <><><><> %
% ~~~ Exception cases ~~~ %
if isempty(posSetA) || isempty(posSetB)
    if isempty(posSetA) && isempty(posSetB)
        posList = [];
        disp('Warning: both "posSetA" and "posSetB" are empty! returning "posList = []"');
        return;
    end
    if isempty(posSetA)
        posList = posSetB;
        disp('Warning: "posSetA" is empty. returning "posList = posSetB"');
    end
    if isempty(posSetB)
        posList = posSetA;
        disp('Warning: "posSetB" is empty. returning "posList = posSetA"'); 
    end
    % > prevent returning non-normalised intensities if exception case occurs
    %   and "params.giveBackNormalisedIntensity" is set to "true".
    if giveNormal && length(posList(1,:))>2
        tmpMn = min(posList(:,3));
        tmpMx = max(posList(:,3));
        tmpRng = tmpMx - tmpMn;
        posList(:,3) = (posList(:,3)-tmpMn)/tmpRng; % < normalise
    end
    return;
end
% ~~~ Exception cases END ~~~ %
% <><><><> Selection of variables and functions <><><><> %
if prioritiseA
    setP1 = posSetA;
    setP1rows = Arows;
    setP2 = posSetB;
    setP2rows = Brows;
else
    setP1 = posSetB;
    setP1rows = Brows;
    setP2 = posSetA;
    setP2rows = Arows;
end
setP1constBools = true(setP1rows,1);
setP2constBools = true(setP2rows,1);
% --- Normalise all intensities prior to processing --- %
% > NOTE: this is not a good way of adding the functionality of 
%   "params.giveBackNormalisedIntensity" and is just bad coding, but should work
if giveNormal && intensityAvailable
    tmpMn = min(setP1(:,3));
    tmpMx = max(setP1(:,3));
    tmpRng = tmpMx - tmpMn;
    setP1(:,3) = (setP1(:,3)-tmpMn)/tmpRng;
    tmpMn = min(setP2(:,3));
    tmpMx = max(setP2(:,3));
    tmpRng = tmpMx - tmpMn;
    setP2(:,3) = (setP2(:,3)-tmpMn)/tmpRng;
    normFun = normFunDud; % <- set to dud as already normalised values.
else
    normFun = normFunReal;
end
% --- Normalise all intensities prior to processing END --- %
if acurateSlowMethod
    excludeLenish = excludeLen^2;
    lenishFun = preciseDist;
    indentFun = preciseInd;
    tempP2list = zeros(setP2rows,1);
else
    excludeLenish = excludeLen;
    lenishFun = fastDist;
    indentFun = fastInd;
    tempP2list = zeros(setP2rows,2);
end
% <><><><> Selection of variables and functions END <><><><> %
tic;
if intensityUsing 
    % ~~~ prep ~~~ %
    % P1
    P1max = max(setP1(:,3));
    P1min = min(setP1(:,3));
    P1range = P1max - P1min;
    setP1tempBools = true(setP1rows,1);
    % P2
    P2max = max(setP2(:,3));
    P2min = min(setP2(:,3));
    P2range = P2max - P2min;
    setP2tempBools = true(setP2rows,1);
    % ~~~ prep END ~~~ %
    % P1
    P1const1 = 1/P1range; % **
    P1const2 = P1const1*P1min; % **
    % P2
    P2const1 = 1/P2range; % **
    P2const2 = P2const1*P2min; % **
    % ~~~ prep END END ~~~ %
    
    for i = 1:setP1rows
        P1normVal = normFun(setP1(i,3), P1min, P1range); % *
        % P1normVal = normFunI(setP1(i,3), P1const1, P1const2); % **
        tempP2list(:,:) = lenishFun(setP1, setP2, i);
        P2tempIndent = indentFun(tempP2list,excludeLenish);
        P2tempMax = max(setP2(P2tempIndent,3));
        P2normVal = normFun(P2tempMax, P2min, P2range); % *
        % P2normVal = normFunI(P2tempMax, P2const1, P2const2); % **
        if P2normVal > P1normVal
            setP1constBools(i,1) = false;
        else
            setP2constBools(P2tempIndent,1) = false;
        end
        % setP2constBools(indentFun(tempP2list,excludeLenish),1) = false;
    end
    newP2set = setP2(setP2constBools(:,1),:);
    newP1set = setP1(setP1constBools(:,1),:);
    posList = [newP1set;newP2set];
%     disp('WARNING: Intensity based selection method has not been implemented yet! Please set "params.useIntensity" to "false".')
    
else
    for i = 1:setP1rows        
        tempP2list(:,:) = lenishFun(setP1, setP2, i);
        setP2constBools(indentFun(tempP2list,excludeLenish),1) = false;
    end
    newP2set = setP2(setP2constBools(:,1),:);
    posList = [setP1;newP2set];
%     return
end
processtime = toc;

if ~supresDisp
    dispTime = sprintf('<interExclude> Position cleaning process time = %d',processtime);
    disp(dispTime);
end

return

end




