% trackPosModV6(posList, maxDisp, params)
%
% ======= Inputs ======= %
% > posList: Time taged co-ordinate list. Output from "getPos" function. 
% > maxDisp: Limiting maximum displacement (in pixels) between consecutive positions
%            in a trace. Once reached or if no traces could be formed during
%            iterations with lower max displacements, then the function stops looking
%            for any traces requireing larger max displacements.
% ----- Optional inputs ----- %
% > params.___: optional parameters for finer control of function.
% >> params.preTreat: Default = 2 | 0 to 2 or true/false | number of pre-treatment iterations.
% >> params.preDisp(1): int | pre-treatment A (run 1) max displacement.
% >> params.preDisp(2): int | pre-treatment B (run 2) max displacement.
% >> params.preThreshold(1): number | pre-treatment A (run 1) average velocity cut off threshold.
% >> params.preThreshold(2): number | pre-treatment B (run 2) average velocity cut off threshold.
% >> params.thresholdConstant: Default = 1.5 | number | factor multiplied by the average trace velocities to calculate velocity cut off thresholds.
% >> params.paraPre.___: Normal/main iteration specific.
% >>> params.paraNorm.mem: int | how many consecutive missing points a trace can continue with.
% >>> params.paraNorm.good: int | threshold for required number of points in a trace.
% >> params.paraPre.___: Pre-treatment iteration specific.
% >>> params.paraPre.mem:^^
% >>> params.paraPre.good:^^
% >> params.paraEnd.___: End extremity iteration specific (i.e. whan maxDisp is relatively large).
% >>> params.paraEnd.mem:^^
% >>> params.paraEnd.good:^^
% >> params.firstDisp: Default = false(i.e. let function decide) | int or false | initial/first normal iteration max displacement.
%
% ======= Out put ======= %
% > allResults: - Uncombined data, i.e. iterations are seperated.
% > completeRaw: - Combined traces from all iterations. In one big list with extra column for trace ID.
%                - No overlaps. longer traces take presedance.
% > completeTraces: - Same data as in "completeRaw" but with points seperated by the trace they belong to.
%
function [completeTraces, completeRaw, allResults] = trackPosModV6(posList, maxDisp, params) 

% ============ Defaults ============ %

threshHoldFactor = 1.5; % > factor multiplied by the average trace velocities to calculate velocity cut off thresholds.

maxDispStepDown = 2; % > amount decreased by in pixels to the max displacement upon failed itteration.
maxDispStepUp = 10; % > amount increased by in pixels to the max displacement upon succesful completion of normal itteration.

% -------- Normal iterations -------- %
memNorm = 5; % > how many consecutive missing points a trace can continue with.
goodNorm = 5; % > threshold for required number of points in a trace.
dimNorm = 2; % > number of dimensions particles are tracked in ?
quietNorm = 1; % > 1/true - supress info display, 0/false - allow info display.
% -------- Pre treatment iterations -------- %
memPre = 5; %2;
goodPre = 8; %10;
dimPre = dimNorm;
quietPre = quietNorm;
% -<>- (A) 1st pre-treatment -<>- %
preTreatA = true; % > true = allow pretreatment, false = skip pretreatment.
preXthresholdA = 0.015; % > velocity threshold used in trace binning (pixels/ms).
preXdispA = 3; % > max displacement (pixels).
% -<>- (B) 2nd pre-treatment -<>- %
preTreatB = true;
preXthresholdB = 0.021;
preXdispB = 5;
% -------- End extremity iterations -------- %
memEnd = 2;
goodEnd = goodNorm;
% ------ %

% ------ determination of normal initial max displacement ------ %

minXpos=min(posList(:,1));
maxXpos=max(posList(:,1));
minYpos=min(posList(:,2));
maxYpos=max(posList(:,2));
Xdist = maxXpos-minXpos; 
Ydist = maxYpos-minYpos;
XYarea = Xdist*Ydist;

frameNum = length(unique(posList(:,3)));
allPoints = length(posList(:,3));
avgPts = allPoints/frameNum;
avgPtDens = avgPts/XYarea;

if avgPtDens > 0.002
    normFirstMaxDisp = 6;
elseif avgPtDens > 0.001
    normFirstMaxDisp = 7;
elseif avgPtDens > 0.0005
    normFirstMaxDisp = 8;
elseif avgPtDens > 0.0003
    normFirstMaxDisp = 10;
elseif avgPtDens > 0.00001
    normFirstMaxDisp = 15;
elseif avgPtDens > 0.00002
    normFirstMaxDisp = 18;
elseif avgPtDens > 0.000005
    normFirstMaxDisp = 22;
elseif avgPtDens > 0.000003
    normFirstMaxDisp = 26;
elseif avgPtDens > 0.000001
    normFirstMaxDisp = 30;
else
    normFirstMaxDisp = 35;
end

% ----------- input parameter sorting ---------- %

if nargin == 3
    if isfield(params,'preTreat')
        preTreatA = false;
        preTreatB = false;
        if isnumeric(params.preTreat)
            if params.preTreat > 0
                preTreatA = true;
            end
            if params.preTreat > 1
                preTreatB = true;
            end
        elseif params.preTreat
            preTreatA = true;
        end    
    end     
    if isfield(params,'preDisp')
        if length(params.preDisp(:)) > 1
            preXdispA = params.preDisp(1);
            preXdispB = params.preDisp(2);
        elseif params.preDisp ~= false
            preXdispA = params.preDisp;
        end
    end
    if isfield(params,'thresholdConstant')
        if params.thresholdConstant ~= false
            threshHoldFactor = params.thresholdConstant;
        end
    end
    if isfield(params,'paraNorm')
        if isfield(params.paraNorm,'mem')
            memNorm = params.paraNorm.mem;
        end
        if isfield(params.paraNorm,'good')
            goodNorm = params.paraNorm.good;
        end
    end
    if isfield(params,'paraPre')
        if isfield(params.paraPre,'mem')
            memPre = params.paraPre.mem;
        end
        if isfield(params.paraPre,'good')
            goodPre = params.paraPre.good;
        end
    end
    if isfield(params,'paraEnd')
        if isfield(params.paraEnd,'mem')
            memEnd = params.paraEnd.mem;
        end
        if isfield(params.paraEnd,'good')
            goodEnd = params.paraEnd.good;
            normFirstMaxDisp = params.firstDisp;
        end
    end
    if isfield(params,'firstDisp')
        normFirstMaxDisp = params.firstDisp;
    end
    if isfield(params,'preThreshold')
        if length(params.preThreshold(:)) > 1
            preXthresholdA = params.preThreshold(1);
            preXthresholdB = params.preThreshold(2);
        elseif params.preThreshold ~= false
            preXthresholdA = params.preThreshold;
        end
    end
end
if preTreatB == true
    if preXdispA > preXdispB
        tempA = preXdispA;
        preXdispA = preXdispB;
        preXdispB = tempA;
    end
end
if preXthresholdA > preXthresholdB
    tempA = preXthresholdA;
    preXthresholdA = preXthresholdB;
    preXthresholdB = tempA;
end

disp('starting max displacement = ')
disp(normFirstMaxDisp);

% -------------- Set Parameters --------------- %
% - - - - %
dwnChnge = maxDispStepDown;
upChnge = maxDispStepUp;
% - - - - %
param.dim = dimNorm;
param.quiet = quietNorm;
% -
param.mem = memNorm;
param.good = goodNorm;
% - - - - %
paramS.dim = dimPre;
paramS.quiet = quietPre;
% -
paramS.mem = memPre;
paramS.good = goodPre;
% - - - - %

%%%%%%%%%%%%%%%% Parameters END %%%%%%%%%%%%%%%%%%

if maxDisp < normFirstMaxDisp
    XdispSt = maxDisp;
else
    XdispSt = normFirstMaxDisp;
end
if preTreatA
    Xdisp = preXdispA;
else
    Xdisp = XdispSt;
end
posListX = posList;
ii = 0;

posListLength = allPoints;

while Xdisp <= maxDisp && posListLength > 0
    ii = ii+1;
    if ii >= 4
        param.mem = memEnd;
        param.good = goodEnd;
    end
    
    if (ii > 3) || (ii > 2 && preTreatB == false) || (ii > 1 && preTreatA == false)
        Xdisp = Xdisp + upChnge;
    elseif ii == 2 && preTreatB == true
        Xdisp = preXdispB;
    elseif ii == 1 && preTreatA == true
        Xdisp = preXdispA;
    else 
        Xdisp = XdispSt;
    end

    contin = -1;
    while contin == -1
       if ii == 1
           paramsA = paramS; 
       else
           paramsA = param; 
       end
       res0 = trackingMod(posListX,Xdisp,paramsA);
       if length(res0(:,1)) == 1
           if res0 == -1
               contin = -1;
               Xdisp = Xdisp - dwnChnge;
           elseif res0 == -2
               contin = 5;
               fianlDISP = Xdisp - upChnge;
               finalIT = ii - 1;
               disp('final max displacment value used was: ');
               disp(fianlDISP);
               disp('final taking iteration was: ');
               disp(finalIT);
               Xdisp = 2*maxDisp;
           else
               contin = 2;
           end
       else
           contin = 2;
       end
    end
    %%%%%%%%%%%%%%%%% trace list building %%%%%%%%%%%%%%%%%
    if contin == 2
        iDx = unique(res0(:,end));
        nelT = numel(iDx);
        tempTrace = cell(nelT,1); % pre-allocating memory
        for k=1:nelT
          tempTrace{k}=res0(res0(:,end) == iDx(k),1:(end-1)); %% require change
        end
        resBox{ii,1} = res0;
        resBox{ii,2} = tempTrace;
        resBox{ii,3} = Xdisp;
        %%%%%%%%%%%%%%%%% trace list building END %%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% trace speed finder %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        traceZ = tempTrace;
        nel = numel(traceZ);
        speedList = zeros(nel,1);
        for k=1:nel;    
            totTime = traceZ{k}(end,5) - traceZ{k}(1,5);
            deltaX = abs(traceZ{k}(2:end,1) - traceZ{k}(1:(end-1),1));
            deltaY = abs(traceZ{k}(2:end,2) - traceZ{k}(1:(end-1),2));
            deltaXY = sqrt(deltaX.^2 + deltaY.^2);
            deltaXYsum = sum(deltaXY);
            velocityA = deltaXYsum/totTime;
            speedList(k) = velocityA; 
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% trace speed finder END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
        if ii == 1 && preTreatA == true
            speed0 = preXthresholdA;
        elseif ii == 2 && preTreatB == true
            speed0 = preXthresholdB;
        else
            speed0 = threshHoldFactor*mean(speedList);
        end
        posListX = ratePurgeV3(tempTrace,posListX,speed0);
        posListLength = length(posListX);
        disp('tracks performed: ');
        disp(ii);
        disp('purged posList length: ');
        disp(posListLength);
    end
end
traceListsNum = numel(resBox(:,1));
if traceListsNum > 1
    joinNum = traceListsNum-1;
    for Ji = 1:joinNum
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if Ji == 1
            raw0 = resBox{Ji,1};
            traces0 = resBox{Ji,2};
        else
            raw0 = newRaw;
            traces0 = newListTraces;
        end
        raw1 = resBox{(Ji+1),1};
        traces1 = resBox{(Ji+1),2};
%         raw1rows = (1:length(raw1(:,4)))';
        raw0rows = (1:length(raw0(:,4)))';
%         dupbs0 = [];
        dupbs0 = zeros(length(raw0(:,4)),1); % prealocate to maximum possible size using zeros
        idX = 0;
        for ids = 1:length(raw1(:,4))
            tempD = raw0rows(raw0(:,4) == raw1(ids,4));
            if ~isempty(tempD)
                idX = idX+1;
                dupbs0(idX,1) = raw0(tempD,6); % raw0 trace number/index.
            end
        end
        dupbs0(~any(dupbs0,2),:) = [];  % delete rows containing only zeros 
        dupbsU0=unique(dupbs0(:,1));
        traceIdx = [];
        traceIdx(:,1) = 1:numel(traces0(:,1));
        newTraces0 = traces0(setdiff(traceIdx,dupbsU0));

        newListTraces = [traces1;newTraces0];
        Nnel=numel(newListTraces(:,1));
        newRaw = [];
        Nst = 1;
        for Ni = 1:Nnel
            tmpLeng = length(newListTraces{Ni}(:,1));
            Ned = (Nst-1) + tmpLeng;
            newRaw(Nst:Ned,1:5) = newListTraces{Ni};
            newRaw(Nst:Ned,6) = Ni;
            Nst = Ned+1;
        end
    end
else
    newListTraces = resBox{1,1};
    newRaw = resBox{1,2};
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
allResults = resBox; %rawLists, traceLists, displacement (in pixels). 
completeTraces = newListTraces;
completeRaw = newRaw;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%% ROUGH BITS AND BOBS %%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%% ROUGH BITS AND BOBS END %%%%%%%%%%%%%%%%%%%%%%%%%%